/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * Utils.hpp
 *
 *  Created on: 2010/11/11
 *      Author: HARK Development Team
 */

#ifndef UTILS_HPP_
#define UTILS_HPP_

#include <complex>
#include <limits>
#include <stdexcept>
#include <valarray>
#include <utility>
#include <vector>

#include <boost/shared_ptr.hpp>

#include <constants.hpp>

/**
 * @file Utils.hpp
 * 便利な機能を提供するユーティリティファイル.
 * このファイルには、どのヘッダ/ソースに収めるのもしっくりこない
 * 便利機能を収めておく。
 */

namespace FD
{
template <typename T> class RCPtr;
template <typename T> class Vector;
}

namespace NSHarkBinaural
{

/**
 * 型Tから参照を取り除く.
 */
template <typename T>
struct RemoveReference
{
    // Tが非参照型ならそのまま
    typedef T type;
};

/**
 * 型Tから参照を取り除く.
 */
template <typename T>
struct RemoveReference<T&>
{
    // Tが参照型なら非参照型にする
    typedef T type;
};

/**
 * 型Tを参照型にする.
 */
template <typename T>
struct AddReference
{
    // Tが非参照型なら参照型にする
    typedef T& type;
};

/**
 * 型Tを参照型にする.
 */
template <typename T>
struct AddReference<T&>
{
    // Tが参照型ならそのまま
    typedef T& type;
};

/**
 * 型Tを非定数型にする.
 */
template <typename T>
struct RemoveConst
{
    typedef T type;
};

/**
 * 型Tを非定数型にする.
 */
template <typename T>
struct RemoveConst<const T>
{
    typedef T type;
};

/**
 * 型Tを定数型にする.
 */
template <typename T>
struct AddConst
{
    typedef const T type;
};

/**
 * 型Tを定数型にする.
 */
template <typename T>
struct AddConst<const T>
{
    typedef const T type;
};

// コンパイル依存を減らすためにクラス名宣言
template <typename, typename> class ChannelHolder;
template <typename> class SignalBase;

/**
 * T型がquiet_NaNを持っている場合のDecibelEnergyの実装.
 */
template <typename T, bool has_quiet_NaN>
struct DecibelEnergyStruct
{
    /**
     * 引数xをデシベル値に変換する.
     * 引数xが正なら、デシベル値を返す。
     * 引数xが非正なら、NaNを返す。
     * @return xのデシベル値
     */
    static T DecibelEnergy(T x)
    {
        return x > 0 ? 10 * log10(x) : std::numeric_limits<T>::quiet_NaN();
    }
};

/**
 * T型がquiet_NaNを持っていない場合のDecibelEnergyの実装.
 */
template <typename T>
struct DecibelEnergyStruct<T, false>
{
    /**
     * 引数xをデシベル値に変換する.
     * 引数xが正なら、デシベル値を返す。
     * 引数xが非正なら、0を返す。
     * @return xのデシベル値
     */
    static T DecibelEnergy(T x)
    {
        return x > 0 ? 10 * log10(x) : 0;
    }
};

/**
 * 引数xをデシベル値に変換する.
 * 引数が、なんらかのエネルギーを表している場合、この関数は期待した値を計算するが、
 * 引数が電圧など、エネルギーではない場合、結果を2倍しなければならない
 * @return xのデシベル値。xが非正の場合は、NaN（NaNが無い環境ならば0）になる。
 */
template <typename T>
inline T DecibelEnergy(T x)
{
    return DecibelEnergyStruct<T, std::numeric_limits<T>::has_quiet_NaN>::DecibelEnergy(x);
}

/**
 * 複素数cのエネルギーのデシベル値を計算する.
 */
template <typename T>
inline T Decibel(std::complex<T> c)
{
    return DecibelEnergy(norm(c));
}

/**
 * std::valarrayからFD::Vectorへ変換する.
 * @param array 変換元の配列
 */
template <typename T>
FD::RCPtr<FD::Vector<T> > ToFDVector(const std::valarray<T>& array)
{
    // FD::Vectorのよくない点を見つけた。
    // FD::Vectorはstd::vectorをpublic継承してしまっている！
    // std::vectorはデストラクタがvirtualではないので、
    //  std::vector<T> *vect = new FD::Vector<T>();
    //  delete vect;
    // ってやるとダメなのだ。
    // 文句言ってもしかたないのでFD::Vectorを使うけどね・・・
    FD::RCPtr<FD::Vector<T> > result(new FD::Vector<T>(array.size()));
    for (size_t i = 0; i < array.size(); ++i)
    {
        result->at(i) = array[i];
    }
    return result;
}

//template <typename T> class SignalBase;

/**
 * 複素数c1に対する複素数c0の相対角度を計算する.
 * @param c0 相対角度を求める複素数
 * @param c1 基準とする複素数
 * @return 相対角度
 */
template <typename T>
T PhaseDifference(std::complex<T> c0, std::complex<T> c1)
{
    // c1に対するc0の角度
    T diff = arg(c0) - arg(c1);

    // -M_PI <= diff <= M_PI に調整
    if (diff > Constants::ConstPi)
    {
        diff -= 2 * Constants::ConstPi;
    }
    else if (diff < -Constants::ConstPi)
    {
        diff += 2 * Constants::ConstPi;
    }

    return diff;
}

/**
 * 1970年1月1日00:00:00 UTCからの経過時間の秒数を得る.
 */
double GetTime();

template <typename T>
std::vector<std::pair<size_t, T> > MakeIndexedVector(const std::vector<T>& v)
{
    std::vector<std::pair<size_t, T> > result;
    result.reserve(v.size());
    for (size_t i = 0; i < v.size(); ++i)
    {
        result.push_back(std::pair<size_t, T>(i, v[i]));
    }
    return result;
}

}

#endif /* UTILS_HPP_ */
